import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

sns.set(style="white", color_codes=True)
import warnings

warnings.filterwarnings("ignore")
from sklearn.linear_model import SGDClassifier
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import BaggingRegressor, RandomForestClassifier, RandomForestRegressor
from sklearn.svm import LinearSVC, LinearSVR
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.linear_model import LassoLarsCV
from sklearn.linear_model import LogisticRegressionCV
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.multiclass import OneVsRestClassifier
from sklearn.neighbors import KNeighborsRegressor
import time
import operator


#########################################################
def QuickML_Ensembling(X_train, y_train, X_test, y_test='', modeltype='Regression',
                       Boosting_Flag=False,
                       scoring='', verbose=0):
    """
    Quickly builds and runs multiple models for a clean data set(only numerics).
    """
    start_time = time.time()
    seed = 99
    FOLDS = 5
    model_dict = {}
    model_tuples = []
    if len(X_train) <= 100000 and X_train.shape[1] < 50:
        NUMS = 100
    else:
        try:
            X_train = X_train.sample(frac=0.30, random_state=99)
            y_train = y_train[X_train.index]
        except:
            pass
        NUMS = 200
    ###### This is where we start performing ensembling of multiple simpler models ###
    if modeltype == 'Regression':
        if scoring == '':
            scoring = 'neg_mean_squared_error'
        # scv = ShuffleSplit(n_splits=FOLDS,random_state=seed)
        scv = KFold(n_splits=FOLDS, shuffle=False)
        if Boosting_Flag is None:
            ## Create an ensemble model ####
            model5 = AdaBoostRegressor(estimator=DecisionTreeRegressor(
                random_state=seed, max_depth=1, min_samples_leaf=2
            ), n_estimators=NUMS, random_state=seed)
            model_tuples.append(('Adaboost', model5))
        elif not Boosting_Flag:
            model5 = LassoLarsCV(cv=scv)
            model_tuples.append(('LassoLarsCV', model5))
        else:
            model5 = LassoLarsCV(cv=scv)
            model_tuples.append(('LassoLarsCV', model5))
        if Boosting_Flag is None:
            model6 = BaggingRegressor(DecisionTreeRegressor(random_state=seed),
                                      n_estimators=NUMS, random_state=seed)
            model_tuples.append(('Bagging_Regressor', model6))
        elif not Boosting_Flag:
            model6 = LinearSVR()
            model_tuples.append(('Linear_SVR', model6))
        else:
            model6 = DecisionTreeRegressor(max_depth=5, min_samples_leaf=2)
            model_tuples.append(('Decision_Tree', model6))
        model7 = KNeighborsRegressor(n_neighbors=8)
        model_tuples.append(('KNN_Regressor', model7))
        if Boosting_Flag is None:
            #### If the Boosting_Flag is True, it means Boosting model is present.
            ###   So choose a different kind of classifier here
            model8 = DecisionTreeRegressor(max_depth=5, min_samples_leaf=2)
            model_tuples.append(('Decision_Tree', model8))
        elif not Boosting_Flag:
            #### If the Boosting_Flag is True, it means Boosting model is present.
            ###   So choose a different kind of classifier here
            model8 = AdaBoostRegressor(base_estimator=DecisionTreeRegressor(
                random_state=seed, max_depth=1, min_samples_leaf=2
            ), n_estimators=NUMS, random_state=seed)
            model_tuples.append(('Adaboost', model8))
        else:
            model8 = RandomForestRegressor(bootstrap=False,
                                           max_depth=10,
                                           max_features='auto',
                                           min_samples_leaf=2,
                                           n_estimators=200,
                                           random_state=99)
            model_tuples.append(('RF_Regressor', model8))
    else:
        if scoring == '':
            scoring = 'accuracy'
        num_classes = len(np.unique(y_test))
        scv = StratifiedKFold(n_splits=FOLDS, shuffle=True, random_state=seed)
        if Boosting_Flag is None:
            ## Create an ensemble model ####
            model5 = AdaBoostClassifier(base_estimator=DecisionTreeClassifier(
                random_state=seed, max_depth=1, min_samples_leaf=2
            ), n_estimators=NUMS, random_state=seed)
            model_tuples.append(('Adaboost', model5))
        elif not Boosting_Flag:
            model5 = LinearDiscriminantAnalysis()
            model_tuples.append(('Linear_Discriminant', model5))
        else:
            model5 = LogisticRegressionCV(Cs=[0.001, 0.01, 0.1, 1, 10, 100], tol=0.01,
                                          solver='liblinear', random_state=seed)
            model_tuples.append(('Logistic_Regression_CV', model5))
        if Boosting_Flag is None:
            model6 = DecisionTreeClassifier(max_depth=5, min_samples_leaf=2)
            model_tuples.append(('Decision_Tree', model6))
        elif not Boosting_Flag:
            model6 = LinearSVC()
            model_tuples.append(('Linear_SVC', model6))
        else:
            model6 = DecisionTreeClassifier(max_depth=5, min_samples_leaf=2)
            model_tuples.append(('Decision_Tree', model6))
        if modeltype == 'Binary_Classification':
            model7 = GaussianNB()
        else:
            model7 = MultinomialNB()
        model_tuples.append(('Naive_Bayes', model7))
        if Boosting_Flag is None:
            #### If the Boosting_Flag is True, it means Boosting model is present.
            ###   So choose a different kind of classifier here
            model8 = RandomForestClassifier(bootstrap=False,
                                            max_depth=10,
                                            max_features='auto',
                                            min_samples_leaf=2,
                                            n_estimators=200,
                                            random_state=99)
            model_tuples.append(('Bagging_Classifier', model8))
        elif not Boosting_Flag:
            #### If the Boosting_Flag is True, it means Boosting model is present.
            ###   So choose a different kind of classifier here
            sgd_best_model = SGDClassifier(alpha=1e-06,
                                           loss='log',
                                           max_iter=1000,
                                           penalty='l2',
                                           learning_rate='constant',
                                           eta0=.1,
                                           random_state=3,
                                           tol=None)
            model8 = OneVsRestClassifier(sgd_best_model)
            model_tuples.append(('One_vs_Rest_Classifier', model8))
        else:
            model8 = RandomForestClassifier(bootstrap=False,
                                            max_depth=10,
                                            max_features='auto',
                                            min_samples_leaf=2,
                                            n_estimators=200,
                                            random_state=99)
            model_tuples.append(('Bagging_Classifier', model8))
    model_dict = dict(model_tuples)
    models, results = run_ensemble_models(model_dict, X_train, y_train, X_test, y_test,
                                          scoring, modeltype)
    return models, results


#########################################################
from sklearn.metrics import balanced_accuracy_score, accuracy_score, precision_score, recall_score, f1_score
from sklearn.metrics import mean_absolute_error, mean_squared_error
import copy
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression


def run_ensemble_models(model_dict, X_train, y_train, X_test, y_test, scoring, modeltype):
    start_time = time.time()
    model_name, bac_score_list, ac_score_list, p_score_list, r_score_list, f1_score_list = [], [], [], [], [], []
    iteration = 0
    estimators = []
    estim_tuples = []
    for key, val in model_dict.items():
        estimator_name = key
        model_name.append(key)
        if str(val).split("(")[0] == 'MultinomialNB':
            #### Multinomial models need only positive values!!
            val.fit(abs(X_train), y_train)
            y_pred = val.predict(abs(X_test))
        else:
            try:
                val.fit(X_train, y_train)
                y_pred = val.predict(X_test)
            except:
                if modeltype != 'Regression':
                    key = 'Logistic_Regression'
                    val = LogisticRegression()
                else:
                    key = 'Linear_Regression'
                    val = LinearRegression()
                val.fit(X_train, y_train)
        try:
            y_pred = val.predict(X_test)
        except:
            print('Error in ensemble models. Returning...')
            stacks = np.array([])
            return model_name, stacks
        if iteration == 0:
            stacks = copy.deepcopy(y_pred)
            iteration += 1
        else:
            stacks = np.c_[stacks, y_pred]
        if not isinstance(y_test, str):
            if modeltype == 'Regression':
                bac_score = np.sqrt(mean_squared_error(y_test, y_pred))
                estimators.append((estimator_name, val, bac_score))
                estim_tuples.append((estimator_name, bac_score))
                bac_score_list.append(bac_score)
                ac_score_list.append(mean_squared_error(y_test, y_pred))
                p_score_list.append(mean_absolute_error(y_test, y_pred))
                model_comparison_df = pd.DataFrame([model_name, bac_score_list, ac_score_list, p_score_list]).T
                model_comparison_df.columns = ['model_name', 'RMSE', 'MSE', 'MAE']
                model_comparison_df = model_comparison_df.sort_values(by='RMSE', ascending=True)
            else:
                bac_score = balanced_accuracy_score(y_test, y_pred)
                estimators.append((estimator_name, val, bac_score))
                estim_tuples.append((estimator_name, bac_score))
                bac_score_list.append(balanced_accuracy_score(y_test, y_pred))
                ac_score_list.append(accuracy_score(y_test, y_pred))
                p_score_list.append(precision_score(y_test, y_pred, average='macro'))
                r_score_list.append(recall_score(y_test, y_pred, average='macro'))
                f1_score_list.append(f1_score(y_test, y_pred, average='macro'))
                model_comparison_df = pd.DataFrame([model_name, bac_score_list, ac_score_list, p_score_list,
                                                    r_score_list, f1_score_list]).T
                model_comparison_df.columns = ['model_name', 'bal_accuracy_score', 'accuracy_score',
                                               'ave_precision_score', 'ave_recall_score', 'ave_f1_score']
                model_comparison_df = model_comparison_df.sort_values(by='bal_accuracy_score', ascending=False)
    if not isinstance(y_test, str):
        data_frame = model_comparison_df.set_index('model_name').astype(float)
        plt.figure(figsize=(10, 10))
        g = sns.heatmap(data_frame, annot=True, fmt='0.2f', cbar=False)
        g.set_xticklabels(g.get_xticklabels(), rotation=45, fontsize=12)
        g.set_yticklabels(g.get_yticklabels(), rotation=0, fontsize=12)
        print('Time taken = %0.0f seconds' % (time.time() - start_time))
        g.set_title('QuickML Ensembling Models Results', fontsize=18)
        f1_stats = dict(estim_tuples)
        try:
            if scoring in ['logloss', 'rmse', 'mae', 'mape', 'RMSE', 'neg_mean_squared_error']:
                best_model_name = min(f1_stats.items(), key=operator.itemgetter(1))[0]
            else:
                best_model_name = max(f1_stats.items(), key=operator.itemgetter(1))[0]
            print('Based on trying multiple models, Best type of algorithm for this data set is %s' % best_model_name)
        except:
            print('Could not detect best algorithm type from ensembling. Continuing...')
    return model_name, stacks
#########################################################



import warnings

import numpy as np
import pandas as pd

warnings.filterwarnings("ignore")
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.svm import LinearSVR
from sklearn.tree import DecisionTreeRegressor
from sklearn.linear_model import LassoCV
from sklearn.model_selection import cross_val_predict
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.tree import DecisionTreeClassifier
import time
import copy
from collections import Counter
from collections import OrderedDict


#############################################################################
def find_rare_class(classes, verbose=0):
    ######### Print the % count of each class in a Target variable  #####
    """
    Works on Multi Class too. Prints class percentages count of target variable.
    It returns the name of the Rare class (the one with the minimum class member count).
    This can also be helpful in using it as pos_label in Binary and Multi Class problems.
    """
    counts = OrderedDict(Counter(classes))
    total = sum(counts.values())
    if verbose >= 1:
        print(' Class  -> Counts -> Percent')
        for cls in counts.keys():
            print("%6s: % 7d  ->  % 5.1f%%" % (cls, counts[cls], counts[cls] / total * 100))
    if type(pd.Series(counts).idxmin()) == str:
        return pd.Series(counts).idxmin()
    else:
        return int(pd.Series(counts).idxmin())


################################################################################
def QuickML_Stacking(X_train, y_train, X_test='', modeltype='Regression', Boosting_Flag=False,
                     scoring='', verbose=0):
    """
    Quickly build Stacks of multiple model results
    Input must be a clean data set (only numeric variables, no categorical or string variables).
    """
    X_train = copy.deepcopy(X_train)
    X_test = copy.deepcopy(X_test)
    y_train = copy.deepcopy(y_train)
    start_time = time.time()
    seed = 99
    if len(X_train) <= 100000 or X_train.shape[1] < 50:
        NUMS = 100
        FOLDS = 5
    else:
        NUMS = 200
        FOLDS = 10
    ## create Stacking models
    estimators = []
    ### This keeps tracks of the number of predict_proba columns generated by each model ####
    estimator_length = []
    if isinstance(X_test, str):
        no_fit = True
    else:
        no_fit = False
    if no_fit:
        #### This is where you don't fit the model but just do cross_val_predict ####
        if modeltype == 'Regression':
            if scoring == '':
                scoring = 'neg_mean_squared_error'
            scv = KFold(n_splits=FOLDS, shuffle=False)
            if Boosting_Flag:
                ######    Bagging models if Bagging is chosen ####
                # model4 = BaggingRegressor(DecisionTreeRegressor(random_state=seed),
                #                            n_estimators=NUMS,random_state=seed)
                model4 = LinearSVR()
                results = cross_val_predict(model4, X_train, y_train, cv=scv, n_jobs=-1)
                estimators.append(('Linear_SVR', model4))
                estimator_length.append(1)
            elif Boosting_Flag is None:
                ####   Tree models if Linear chosen #####
                model5 = DecisionTreeRegressor(random_state=seed, min_samples_leaf=2)
                results = cross_val_predict(model5, X_train, y_train, cv=scv, n_jobs=-1)
                estimators.append(('Decision Trees', model5))
                estimator_length.append(1)
            else:
                ####   Linear Models if Boosting is chosen #####
                model6 = LassoCV(alphas=np.logspace(-5, -1, 20), cv=scv, random_state=seed)
                results = cross_val_predict(model6, X_train, y_train, cv=scv, n_jobs=-1)
                estimators.append(('LassoCV Regularization', model6))
                estimator_length.append(1)
        else:
            n_classes = len(Counter(y_train))
            if n_classes > 2:
                #### In multi-class setting, it makes sense to turn it into binary class in stage-1
                #### In stage 2, a complex model will take the inputs of this model and try to predict
                rare_class = find_rare_class(y_train)
                if rare_class == 0:
                    majority_class = 1
                else:
                    majority_class = 0
                y_train = y_train.map(lambda x: rare_class if x == rare_class else majority_class)
            if scoring == '':
                scoring = 'accuracy'
            scv = StratifiedKFold(n_splits=FOLDS, random_state=seed, shuffle=True)
            if Boosting_Flag:
                ####   Linear Models if Boosting is chosen #####
                model4 = LinearDiscriminantAnalysis()
                results = cross_val_predict(model4, X_train, y_train, cv=scv, n_jobs=-1,
                                            method='predict_proba')
                estimators.append(('Linear Discriminant', model4))
                estimator_length.append(results.shape[1])
            elif Boosting_Flag is None:
                ####   Tree models if Linear chosen #####
                model6 = DecisionTreeClassifier(min_samples_leaf=2)
                results = cross_val_predict(model6, X_train, y_train, cv=scv, n_jobs=-1,
                                            method='predict_proba')
                estimators.append(('Decision Tree', model6))
                estimator_length.append(results.shape[1])
            else:
                ######    Naive Bayes models if Bagging is chosen ####
                if n_classes <= 2:
                    try:
                        model7 = GaussianNB()
                    except:
                        model7 = DecisionTreeClassifier(min_samples_leaf=2)
                else:
                    try:
                        model7 = MultinomialNB()
                    except:
                        model7 = DecisionTreeClassifier(min_samples_leaf=2)
                results = cross_val_predict(model7, X_train, y_train, cv=scv, n_jobs=-1,
                                            method='predict_proba')
                estimators.append(('Naive Bayes', model7))
                estimator_length.append(results.shape[1])
    else:
        #### This is where you fit the model and then predict ########
        if modeltype == 'Regression':
            if scoring == '':
                scoring = 'neg_mean_squared_error'
            scv = KFold(n_splits=FOLDS, shuffle=False)
            if Boosting_Flag:
                ######    Bagging models if Bagging is chosen ####
                # model4 = BaggingRegressor(DecisionTreeRegressor(random_state=seed),
                #                            n_estimators=NUMS,random_state=seed)
                model4 = LinearSVR()
                results = model4.fit(X_train, y_train).predict(X_test)
                estimators.append(('Linear_SVR', model4))
                estimator_length.append(1)
            elif Boosting_Flag is None:
                ####   Tree models if Linear chosen #####
                model5 = DecisionTreeRegressor(random_state=seed, min_samples_leaf=2)
                results = model5.fit(X_train, y_train).predict(X_test)
                estimators.append(('Decision Trees', model5))
                estimator_length.append(1)
            else:
                ####   Linear Models if Boosting is chosen #####
                model6 = LassoCV(alphas=np.logspace(-5, -1, 20), cv=scv, random_state=seed)
                results = model6.fit(X_train, y_train).predict(X_test)
                estimators.append(('LassoCV Regularization', model6))
                estimator_length.append(1)
        else:
            n_classes = len(Counter(y_train))
            if n_classes > 2:
                #### In multi-class setting, it makes sense to turn it into binary class in stage-1
                #### In stage 2, a complex model will take the inputs of this model and try to predict
                rare_class = find_rare_class(y_train)
                if rare_class == 0:
                    majority_class = 1
                else:
                    majority_class = 0
                y_train = y_train.map(lambda x: rare_class if x == rare_class else majority_class)
            if scoring == '':
                scoring = 'accuracy'
            scv = StratifiedKFold(n_splits=FOLDS, random_state=seed, shuffle=True)
            if Boosting_Flag:
                ####   Linear Models if Boosting is chosen #####
                model4 = LinearDiscriminantAnalysis()
                results = model4.fit(X_train, y_train).predict_proba(X_test)
                estimators.append(('Linear Discriminant', model4))
                estimator_length.append(results.shape[1])
            elif Boosting_Flag is None:
                ####   Tree models if Linear chosen #####
                model6 = DecisionTreeClassifier(min_samples_leaf=2)
                results = model6.fit(X_train, y_train).predict_proba(X_test)
                estimators.append(('Decision Tree', model6))
                estimator_length.append(results.shape[1])
            else:
                ######    Naive Bayes models if Bagging is chosen ####
                if n_classes <= 2:
                    try:
                        model7 = GaussianNB()
                    except:
                        model7 = DecisionTreeClassifier(min_samples_leaf=2)
                else:
                    try:
                        model7 = MultinomialNB()
                    except:
                        model7 = DecisionTreeClassifier(min_samples_leaf=2)
                results = model7.fit(X_train, y_train).predict_proba(X_test)
                estimators.append(('Naive Bayes', model7))
                estimator_length.append(results.shape[1])
    # stacks = np.c_[results1,results2,results3]
    estimators_list = [(tuples[0], tuples[1]) for tuples in estimators]
    estimator_names = [tuples[0] for tuples in estimators]
    #### Here is where we consolidate the estimator names and their results into one common list ###
    ls = []
    for x, y in dict(zip(estimator_names, estimator_length)).items():
        els = [x + '_' + str(eachy) for eachy in range(y)]
        ls += els
    if verbose == 1:
        print('    Time taken for Stacking: %0.1f seconds' % (time.time() - start_time))
    return ls, results
#########################################################



import warnings

import numpy as np
from sklearn.metrics import accuracy_score, confusion_matrix, f1_score, fbeta_score, log_loss
from sklearn.metrics import mean_squared_error, median_absolute_error, mean_absolute_error
from sklearn.metrics import mean_squared_log_error, precision_score, recall_score, roc_auc_score


def balanced_accuracy_score(y_true, y_pred, sample_weight=None,
                            adjusted=False):
    matrix = confusion_matrix(y_true, y_pred, sample_weight=sample_weight)
    with np.errstate(divide='ignore', invalid='ignore'):
        per_class = np.diag(matrix) / matrix.sum(axis=1)
    if np.any(np.isnan(per_class)):
        warnings.warn('y_pred contains classes not in y_true')
        per_class = per_class[~np.isnan(per_class)]
    score = np.mean(per_class)
    if adjusted:
        n_classes = len(per_class)
        chance = 1 / n_classes
        score -= chance
        score /= 1 - chance
    return score


#####################################################################################
def accu(results, y_cv):
    return (results == y_cv).astype(int).sum(axis=0) / (y_cv.shape[0])


def rmse(results, y_cv):
    return np.sqrt(np.mean((results - y_cv) ** 2, axis=0))


# Defining objective functions for HyperOpt here
def gini(truth, predictions):
    g = np.asarray(np.c_[truth, predictions, np.arange(len(truth))], dtype=np.float)
    g = g[np.lexsort((g[:, 2], -1 * g[:, 1]))]
    gs = g[:, 0].cumsum().sum() / g[:, 0].sum()
    gs -= (len(truth) + 1) / 2.
    return gs / len(truth)


def gini_sklearn(truth, predictions):
    return gini(truth, predictions) / gini(truth, truth)


def gini_meae(truth, predictions):
    score = median_absolute_error(truth, predictions)
    return score


def gini_msle(truth, predictions):
    score = mean_squared_log_error(truth, predictions)
    return score


def gini_mae(truth, predictions):
    score = mean_absolute_error(truth, predictions)
    return score


def gini_mse(truth, predictions):
    score = mean_squared_error(truth, predictions)
    return score


def gini_rmse(truth, predictions):
    score = np.sqrt(mean_squared_error(truth, predictions))
    return score


def gini_accuracy(truth, predictions):
    return accuracy_score(truth, predictions)


def gini_bal_accuracy(truth, predictions):
    return balanced_accuracy_score(truth, predictions)


def gini_roc(truth, predictions):
    return roc_auc_score(truth, predictions)


def gini_precision(truth, predictions, pos_label=1):
    return precision_score(truth, predictions, average=None)[pos_label]


def gini_average_precision(truth, predictions):
    return precision_score(truth, predictions.argmax(axis=1), average='macro')


def gini_weighted_precision(truth, predictions):
    return precision_score(truth, predictions.argmax(axis=1), average='weighted')


def gini_macro_precision(truth, predictions):
    return precision_score(truth, predictions.argmax(axis=1), average='macro')


def gini_micro_precision(truth, predictions):
    return precision_score(truth, predictions.argmax(axis=1), average='micro')


def gini_samples_precision(truth, predictions):
    return precision_score(truth, predictions.argmax(axis=1), average='samples')


def gini_f1(truth, predictions, pos_label=1):
    return f1_score(truth, predictions, average=None)[pos_label]


def gini_weighted_f1(truth, predictions):
    return f1_score(truth, predictions.argmax(axis=1), average='weighted')


def gini_macro_f1(truth, predictions):
    return f1_score(truth, predictions.argmax(axis=1), average='macro')


def gini_micro_f1(truth, predictions):
    return f1_score(truth, predictions.argmax(axis=1), average='micro')


def gini_samples_f1(truth, predictions):
    return f1_score(truth, predictions.argmax(axis=1), average='samples')


def gini_log_loss(truth, predictions):
    return log_loss(truth, predictions, normalize=True)


def gini_recall(truth, predictions, pos_label=1):
    return recall_score(truth, predictions, average=None)[pos_label]


def gini_weighted_recall(truth, predictions):
    return recall_score(truth, predictions.argmax(axis=1), average='weighted')


def gini_samples_recall(truth, predictions):
    return recall_score(truth, predictions.argmax(axis=1), average='samples')


def gini_macro_recall(truth, predictions):
    return recall_score(truth, predictions.argmax(axis=1), average='macro')


def gini_micro_recall(truth, predictions):
    return recall_score(truth, predictions.argmax(axis=1), average='micro')


# calculate f2-measure
def f2_measure(y_true, y_pred):
    return fbeta_score(y_true, y_pred, beta=2)



import pandas as pd
import numpy as np

######### NEW And FAST WAY to ADD Feature Engg COLUMNS IN A DATA SET #######
###   Modify Dataframe by adding Features using Feature Tools ####
def add_computational_primitive_features(df, add_types: list, idcolumn=''):
    """
    ###   Modify Dataframe by adding computational primitive Features using Feature Tools ####
    ###   What are computational primitives? they are to add, subtract, multiply and divide features
    ### Inputs:
    ###   df: Just sent in the data frame df that you want features added to
    ###   add_types: list of computational types: 'add', 'subtract', 'multiply' and 'divide'. Choose any or all.
    ###   idcolumn: this is to create an index for the dataframe since FT runs on index variable. You can leave it empty string.
    """
    try:
        import featuretools as ft
    except:
        print('pip install featuretools and try this again')
    df = copy.deepcopy(df)
    projectid = 'project_prediction'
    dataid = 'project_data'
    if idcolumn == '':
        indexid = 'index'
        make_index = True
    else:
        indexid = idcolumn
        make_index = False
    # Make an entityset and add the entity
    es = ft.EntitySet(id=projectid)
    es.entity_from_dataframe(entity_id=dataid, dataframe=df,
                             make_index=make_index, index=indexid)

    # Run deep feature synthesis with given input primitives or automatically deep 2
    if len(add_types) > 0:
        ### Build Features based on given primitive types, add_types which is a list
        df_mod, feature_defs = ft.dfs(entityset=es, target_entity=dataid,
                                      trans_primitives=add_types)
    else:
        ### Perform Deep Feature Synthesis Automatically for Depth 2
        df_mod, feature_defs = ft.dfs(entityset=es, target_entity=dataid,
                                      max_depth=2, n_jobs=-1,
                                      verbose=0)
    if make_index:
        df_mod = df_mod.reset_index(drop=True)
    return df_mod


def feature_engineering(df, ft_requests, idcol):
    """
    The Feature Engineering module needs FeatureTools installed to work.
    So please do "pip install featuretools" before trying out this module.
    It takes a given data set, df and adds features based on the requet types in
    ft_requests which can be 'add','subtract','multiply','divide'. If you have
    an id_column in the data set, you can provide it as idcol (a string variable).
    It will return your modified dataset with 'idcol' as the index. Make sure
    you reset the index if you want to return it to its former state.
    Also do not send in your entire dataframe! Just send a small dataframe with a few variables.
    Once you see how it adds to performance of model, you can add more variables to dataframe.
    """
    df = copy.deepcopy(df)
    if df.shape[1] < 2:
        print('More than one column in dataframe required to perform feature engineering. Returning')
        return df
    ft_dict = dict(zip(['add', 'multiply', 'subtract', 'divide'],
                       ['add_numeric', 'multiply_numeric',
                        'subtract_numeric', 'divide_numeric']))
    if len(ft_requests) > 0:
        ft_list = []
        for ft_one in ft_requests:
            if ft_one in ft_dict.keys():
                ft_list.append(ft_dict[ft_one])
            else:
                print('    Cannot perform %s-type feature engineering...' % ft_one)
        cols = [x for x in df.columns.tolist() if x not in [idcol]]
        for each_ft, count in zip(ft_list, range(len(ft_list))):
            if count == 0:
                df_mod = add_computational_primitive_features(df, [each_ft], idcol)
                print(df_mod.shape)
            else:
                df_temp = add_computational_primitive_features(df, [each_ft], idcol)
                df_temp.drop(cols, axis=1, inplace=True)
                df_mod = pd.concat([df_mod, df_temp], axis=1, ignore_index=False)
                print(df_mod.shape)
    else:
        df_mod = add_computational_primitive_features(df, [], idcol)
    return df_mod


def add_date_time_features(smalldf, startTime, endTime, splitter_date_string="/", splitter_hour_string=":"):
    """
    If you have start date time stamp and end date time stamp, this module will create additional features for such fields.
    You must provide a start date time stamp field and if you have an end date time stamp field, you must use it.
    Otherwise, you are better off using the create_date_time_features module which is also in this library.
    You must provide the following:
    smalldf: Dataframe containing your date time fields
    startTime: this is hopefully a string field which converts to a date time stamp easily. Make sure it is a string.
    endTime: this also must be a string field which converts to a date time stamp easily. Make sure it is a string.
    splitter_date_string: usually there is a string such as '/' or '.' between day/month/year etc. Default is assumed / here.
    splitter_hour_string: usually there is a string such as ':' or '.' between hour:min:sec etc. Default is assumed : here.
    """
    smalldf = smalldf.copy()
    add_cols = []
    start_date = 'processing' + startTime + '_start_date'
    smalldf[start_date] = smalldf[startTime].map(lambda x: x.split(" ")[0])
    add_cols.append(start_date)
    try:
        start_time = 'processing' + startTime + '_start_time'
        smalldf[start_time] = smalldf[startTime].map(lambda x: x.split(" ")[1])
        add_cols.append(start_time)
    except:
        ### there is no hour-minutes part of this date time stamp field. You can just skip it if it is not there
        pass
    end_date = 'processing' + endTime + '_end_date'
    smalldf[end_date] = smalldf[endTime].map(lambda x: x.split(" ")[0])
    add_cols.append(end_date)
    try:
        end_time = 'processing' + endTime + '_end_time'
        smalldf[end_time] = smalldf[endTime].map(lambda x: x.split(" ")[1])
        add_cols.append(end_time)
    except:
        ### there is no hour-minutes part of this date time stamp field. You can just skip it if it is not there
        pass
    view_days = 'processing' + startTime + '_elapsed_days'
    smalldf[view_days] = (pd.to_datetime(smalldf[end_date]) - pd.to_datetime(smalldf[start_date])).values.astype(int)
    add_cols.append(view_days)
    try:
        view_time = 'processing' + startTime + '_elapsed_time'
        smalldf[view_time] = (pd.to_datetime(smalldf[end_time]) - pd.to_datetime(smalldf[start_time])).astype(
            'timedelta64[s]').values
        add_cols.append(view_time)
    except:
        ### In some date time fields this gives an error so skip it in that case
        pass
    #### The reason we chose endTime here is that startTime is usually taken care of by another library. So better to do this alone.
    year = 'processing' + endTime + '_end_year'
    smalldf[year] = smalldf[end_date].map(lambda x: str(x).split(splitter_date_string)[0]).values
    add_cols.append(year)
    #### The reason we chose endTime here is that startTime is usually taken care of by another library. So better to do this alone.
    month = 'processing' + endTime + '_end_month'
    smalldf[month] = smalldf[end_date].map(lambda x: str(x).split(splitter_date_string)[1]).values
    add_cols.append(month)
    try:
        #### The reason we chose endTime here is that startTime is usually taken care of by another library. So better to do this alone.
        daynum = 'processing' + endTime + '_end_day_number'
        smalldf[daynum] = smalldf[end_date].map(lambda x: str(x).split(splitter_date_string)[2]).values
        add_cols.append(daynum)
    except:
        ### In some date time fields the day number is not there. If not, just skip it ####
        pass
    #### In some date time fields, the hour and minute is not there, so skip it in that case if it errors!
    try:
        start_hour = 'processing' + startTime + '_start_hour'
        smalldf[start_hour] = smalldf[start_time].map(lambda x: str(x).split(splitter_hour_string)[0]).values
        add_cols.append(start_hour)
        start_min = 'processing' + startTime + '_start_hour'
        smalldf[start_min] = smalldf[start_time].map(lambda x: str(x).split(splitter_hour_string)[1]).values
        add_cols.append(start_min)
    except:
        ### If it errors, skip it
        pass
    #### Check if there is a weekday and weekends in date time columns using endTime only
    weekday_num = 'processing' + endTime + '_end_weekday_number'
    smalldf[weekday_num] = pd.to_datetime(smalldf[end_date]).dt.weekday.values
    add_cols.append(weekday_num)
    weekend = 'processing' + endTime + '_end_weekend_flag'
    smalldf[weekend] = smalldf[weekday_num].map(lambda x: 1 if x in [5, 6] else 0)
    add_cols.append(weekend)
    #### If everything works well, there should be 13 new columns added by module. All the best!
    print('%d columns added using start date=%s and end date=%s processing...' % (len(add_cols), startTime, endTime))
    return smalldf

#################################################################################
def split_one_field_into_many(df, field, splitter, filler, new_names_list):
    """
    This little function takes any data frame field (string variables only) and splits
    it into as many fields as you want in the new_names_list.
    You can also specify what string to split on using the splitter argument.
    You can also fill Null values that occur due to your splitting by specifying a filler.
    if no new_names_list is given, then we use the name of the field itself to split.
    """
    import warnings
    warnings.filterwarnings("ignore")
    df = df.copy()
    ### First print the maximum number of things in that field
    max_things = df[field].map(lambda x: len(x.split(splitter))).max()
    if len(new_names_list) == 0:
        print('    Max. columns created by splitting %s field is %d.' % (
            field, max_things))
    else:
        print(
            '    Max. columns created by splitting %s field is %d but you have given %d variable names only. Selecting first %d' % (
                field, max_things, len(new_names_list), len(new_names_list)))
    ### This creates a new field that counts the number of things that are in that field.
    num_products_viewed = 'count_things_in_' + field
    df[num_products_viewed] = df[field].map(lambda x: len(x.split(";"))).values
    ### Clean up the field such that it has the right number of split chars otherwise add to it
    df[field] = df[field].map(
        lambda x: x + splitter * (max_things - len(x.split(";"))) if len(x.split(";")) < max_things else x)
    ###### Now you create new fields by split the one large field ########
    if new_names_list == '':
        new_names_list = [field + '_' + str(i) for i in range(1, max_things + 1)]
    try:
        for i in range(len(new_names_list)):
            df[field].fillna(filler, inplace=True)
            df.loc[df[field] == splitter, field] = filler
            df[new_names_list[i]] = df[field].map(lambda x: x.split(splitter)[i]
            if splitter in x else x)
    except:
        ### Check if the column is a string column. If not, give an error message.
        print('Cannot split the column. Getting an error. Check the column again')
        return df
    return df, new_names_list


#################################################################################
def add_aggregate_primitive_features(dft, agg_types, id_column, ignore_variables: list):
    """
    ###   Modify Dataframe by adding computational primitive Features using Feature Tools ####
    ###   What are aggregate primitives? they are to "mean""median","mode","min","max", etc. features
    ### Inputs:
    ###   df: Just sent in the data frame df that you want features added to
    ###   agg_types: list of computational types: 'mean','median','count', 'max', 'min', 'sum', etc.
    ###         One caveat: these agg_types must be found in the agg_func of numpy or pandas groupby statement.
    ###         for example: numpy has 'median','prod','sum','std','var', etc. - they will work!
    ###   idcolumn: this is to create an index for the dataframe since FT runs on index variable. You can leave it empty string.
    ###   ignore_variables: list of variables to ignore among numeric variables in data since they may be ID variables.
    """
    import copy
    ### Make sure the list of functions they send in are acceptable functions. If not, the aggregate will blow up!
    func_set = {'count', 'sum', 'mean', 'mad', 'median', 'min', 'max', 'mode', 'abs', 'prod', 'std', 'var', 'sem',
                'skew', 'kurt', 'quantile', 'cumsum', 'cumprod', 'cummax', 'cummin'}
    agg_types = list(set(agg_types).intersection(func_set))
    ### If the ignore_variables list is empty, make sure you add the id_column to it so it can be dropped from aggregation.
    if len(ignore_variables) == 0:
        ignore_variables = [id_column]
    ### Select only integer and float variables to do this aggregation on. Be very careful if there are too many vars.
    ### This will take time to run in that case.
    dft_index = copy.deepcopy(dft[id_column])
    dft_cont = copy.deepcopy(dft.select_dtypes('number').drop(ignore_variables, axis=1))
    dft_cont[id_column] = dft_index
    try:
        dft_full = dft_cont.groupby(id_column).agg(agg_types)
    except:
        ### if for some reason, the groupby blows up, then just return the dataframe as is - no changes!
        return dft
    cols = [x + '_' + y + '_by_' + id_column for (x, y) in dft_full.columns]
    dft_full.columns = cols
    ###  Not every column has useful values. If it is full of just the same value, remove it
    _, list_unique_col_ids = np.unique(dft_full, axis=1, return_index=True)
    dft_full = dft_full.iloc[:, list_unique_col_ids]
    return dft_full


################################################################################################################################
import copy
import time
import pdb


def FE_create_groupby_features(dft, groupby_columns, numeric_columns, agg_types):
    """
    FE means FEATURE ENGINEERING - That means this function will create new features
    Beware: this function will return a smaller dataframe than what you send in since it groups rows by keys.
    #########################################################################################################
    Function groups rows in a dft dataframe by the groupby_columns and returns multiple columns for the numeric column aggregated.
    Do not send in more than one column in the numeric column since beyond the first column it will be ignored!
    agg_type can be any numpy function such as mean, median, sum, count, etc.
    ##########################################################################################################
    Returns: a smaller dataframe with rows grouped by groupby_columns and aggregated for the numeric_column
    """
    start_time = time.time()
    print('Autoviml Feature Engineering: creating groupby features using %s' % groupby_columns)
    ##########  This is where we create new columns by each numeric column grouped by group-by columns given.
    if isinstance(numeric_columns, list):
        pass
    elif isinstance(numeric_columns, str):
        numeric_columns = [numeric_columns]
    else:
        print('    Numeric column must be a string not a number Try again')
        return pd.DataFrame()
    grouped_list = pd.DataFrame()
    for iteration, numeric_column in zip(range(len(numeric_columns)), numeric_columns):
        grouped = dft.groupby(groupby_columns)[[numeric_column]]
        try:
            agg_type = agg_types[iteration]
        except:
            print('    No aggregation type given, hence mean is chosen by default')
            agg_type = 'mean'
        try:
            prefix = numeric_column + '_'
            if agg_type in ['Sum', 'sum']:
                grouped_agg = grouped.sum()
            elif agg_type in ['Mean', 'mean', 'Average', 'average']:
                grouped_agg = grouped.mean()
            elif agg_type in ['count', 'Count']:
                grouped_agg = grouped.count()
            elif agg_type in ['Median', 'median']:
                grouped_agg = grouped.median()
            elif agg_type in ['Maximum', 'maximum', 'max', 'Max']:
                ## maximum of the amounts
                grouped_agg = grouped.max()
            elif agg_type in ['Minimum', 'minimum', 'min', 'Min']:
                ## maximum of the amounts
                grouped_agg = grouped.min()
            else:
                grouped_agg = grouped.mean()
            grouped_sep = grouped_agg.unstack().add_prefix(prefix).fillna(0)
        except:
            print('    Error in creating groupby features...returning with null dataframe')
            grouped_sep = pd.DataFrame()
        if iteration == 0:
            grouped_list = copy.deepcopy(grouped_sep)
        else:
            grouped_list = pd.concat([grouped_list, grouped_sep], axis=1)
        print(
            '    After grouped features added by %s, number of columns = %d' % (numeric_column, grouped_list.shape[1]))
    #### once everything is done, you can close it here
    print('Time taken for creation of groupby features (in seconds) = %0.0f' % (time.time() - start_time))
    try:
        grouped_list.columns = grouped_list.columns.get_level_values(1)
        grouped_list.columns.name = None  ## make sure the name on columns is removed
        grouped_list = grouped_list.reset_index()  ## make sure the ID column comes back
    except:
        print('   Error in setting column names. Please reset column names after this step...')
    return grouped_list
################################################################################
from scipy.stats import skew, kurtosis
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.metrics import mean_squared_error, accuracy_score

class EntropyBinningTransformer(BaseEstimator, TransformerMixin):
    """
    #########   This innovative binning transformer is found only in AutoViML  ##############################
    The entropy binning approach in AutoViML is a novel and intriguing approach, especially as 
    it leverages decision trees to determine optimal binning thresholds based on information gain 
    (for classification) or mean squared error reduction (for regression). 

    This method can adapt well to non-linear relationships between the feature and the target. 
    It also incorporates some heuristics and improvements to enhance its effectiveness and efficiency.

    Heuristics:
    Variable Selection: We prioritize binning the most important variables. 
    This is a sensible heuristic as it focuses computational resources on variables that are 
    most likely to impact the model. We introduce a parameter to specify the number of top variables to bin.

    Adaptive max_depth: We've used an adaptive max_depth based on the number of continuous variables. 
    This heuristic helps control the granularity of the binning. We can maintain this by adding a 
    logic to adjust max_depth based on the number of features.

    Handling of Binned Variables: The code has an option to either keep the original variables 
    alongside the binned versions or replace them. This flexibility can be useful and we keep it 
    in the class implementation.

    Post-Binning Smoothing: After binning, applying smoothing techniques (e.g., Laplace smoothing) 
    can help address bins with very low frequencies, which might be particularly useful for categorical targets.

    Incorporating these aspects, this is an updated version of the original EntropyBinningTransformer method in AutoViML

    Additional Improvements to consider:

    Evaluation and Optimization: Integrating a mechanism to evaluate the effectiveness of binning (e.g., 
        using cross-validation scores) and automatically optimize the binning parameters accordingly.
    #########################################################################################################
    EntropyBinningTransformer is a transformer class for binning numeric variables based on entropy using decision trees,
    with options for selecting the most relevant variables for binning based on their predictive power, skewness,
    and kurtosis, and for smoothing the bins post-binning to mitigate issues with overfitting and bins with 
        low frequencies.

    Parameters:
    - max_depth: The maximum depth of the decision trees used for binning.
    - min_samples_leaf: The minimum number of samples required to be at a leaf node of the decision trees.
    - entropy_binning: Whether to apply entropy binning.
    - modeltype: Indicates whether the target variable is for 'Regression' or 'Classification'.

    Methods:
    - fit: Learns the binning thresholds for each variable.
    - transform: Applies the learned thresholds to bin the data, with an option for post-binning smoothing.
    - select_top_n_vars: Selects the top N variables for binning based on their predictive power and distribution
         characteristics.
    #########################################################################################################
    ####  U S A G E  #####
    # Initialize the EntropyBinningTransformer 
    entropy_binner = EntropyBinningTransformer( replace_vars=True, modeltype='Classification')

    # Optionally, select top n variables based on their predictive power
    # This step is useful if you want to bin only the most informative variables
    top_vars = entropy_binner.select_top_n_vars(X_train, y_train, n=2)
    X_train_top_vars = X_train[top_vars]

    # Fit the transformer to the training data
    entropy_binner.fit(X_train_top_vars, y_train)

    # Transform the training data and new data using the learned binning thresholds
    X_train_binned = entropy_binner.transform(X_train_top_vars)
    X_test_binned = entropy_binner.transform(X_test[top_vars])    
    #########################################################################################################
    """
    def __init__(self, replace_vars=True, modeltype='Classification', top_n_vars=None):
        self.replace_vars = replace_vars
        self.modeltype = modeltype
        self.max_depth = 10
        self.min_samples_leaf = 2
        self.top_n_vars = top_n_vars
        self.binning_thresholds = {}
        self.remvars = []

    def fit(self, X, y):
        # Determine top_n variables based on importance, for example, using feature importances from a preliminary model
        if self.top_n_vars is not None:
            top_vars = self.select_top_n_vars(X, y, n=self.top_n_vars)
        else:
            top_vars = X.columns

        for col in top_vars:
            max_depth = self._adjust_max_depth(len(top_vars))
            clf = self.get_decision_tree(max_depth)
            
            try:
                clf.fit(X[col].values.reshape(-1, 1), y)
                thresholds = clf.tree_.threshold[clf.tree_.threshold > -2]
                self.binning_thresholds[col] = np.sort(thresholds)
            except Exception as e:
                print(f'Error in {col} during Entropy Binning: {e}')
        
        return self
        
    def get_decision_tree(self, max_depth):
        if self.modeltype == 'Regression':
            return DecisionTreeRegressor(criterion='mse', min_samples_leaf=self.min_samples_leaf,
                                         max_depth=max_depth, random_state=99)
        else:
            return DecisionTreeClassifier(criterion='entropy', min_samples_leaf=self.min_samples_leaf,
                                          max_depth=max_depth, random_state=99)

    def _adjust_max_depth(self, n_features):
        """
        Adjust max depth based on the heuristic from the original function.

        Future:
        Dynamic Depth Adjustment: Instead of fixed depth adjustments based on the number of variables, 
        use a more dynamic approach that considers the actual distribution and variability of each variable.

        """
        if self.max_depth is not None:
            return self.max_depth  # Use user-defined max_depth if provided
        elif n_features <= 2:
            return 2
        elif n_features <= 5:
            return n_features - 2
        elif n_features <= 10:
            return 5
        else:
            return 10  # Default max depth for > 10 features

    def select_top_n_vars(self, X, y, n=2, skew_threshold=1.5, kurtosis_threshold=3):
        """
        The select_top_n_vars method takes into account the skewness and kurtosis of each variable
         in addition to its predictive power. The final score is used to rank the variables 
         adjusted by the absolute values of skewness and kurtosis to prefer variables that are 
         both predictive and have less extreme distributions.

        Incorporating distribution-based metrics such as skewness, kurtosis, and outlier 
        detection into the selection process for variables to bin can indeed provide a more 
        nuanced approach. Variables with high skewness or kurtosis may benefit more from binning, 
        as it can help to normalize their effect in predictive models. Additionally, handling 
        outliers before binning can ensure that bins are more representative of the general 
        distribution of the data.
        """
        scores = {}
        ### Select only float numeric columns to bin ###########
        numvars = X.select_dtypes(include='float').columns.tolist()
        ### Let's keep the remaining vars ######
        self.remvars = [x for x in list(X) if x not in numvars ]
        for col in numvars:
            X_col = X[col].values.reshape(-1, 1)

            # Calculate skewness and kurtosis
            skewness = skew(X[col])
            kurt = kurtosis(X[col], fisher=True)  # Fisher's definition is used (normal ==> 0.0)

            # Initialize a simple decision tree
            if self.modeltype == 'Regression':
                model = DecisionTreeRegressor(max_depth=self.max_depth, random_state=99)
            else:
                model = DecisionTreeClassifier(max_depth=self.max_depth, random_state=99)

            # Fit the model
            model.fit(X_col, y)

            # Predict using the model
            predictions = model.predict(X_col)

            # Calculate the performance metric
            if self.modeltype == 'Regression':
                score = -mean_squared_error(y, predictions)  # Negative because lower MSE is better
            else:
                score = accuracy_score(y, predictions)

            # Store the score, possibly adjusted for skewness and kurtosis
            if abs(skewness) > skew_threshold or abs(kurt) > kurtosis_threshold:
                scores[col] = score * (1 + abs(skewness) + abs(kurt))
            else:
                scores[col] = score

        # Sort the variables based on the scores and select the top n
        top_n_vars = sorted(scores, key=scores.get, reverse=True)[:max(n, int(X.shape[1] * 0.1), 2)]

        self.remvars += [x for x in numvars if x not in top_n_vars ]
        return top_n_vars

    def transform(self, X):
        """
        Applies the learned binning thresholds and Laplace smoothing to the data, replacing or appending the original 
        variables with their smoothed, binned versions based on the replace_vars flag.

        Parameters:
        - X (pd.DataFrame): The input data to transform.

        Returns:
        - X_transformed (pd.DataFrame): The transformed data with variables binned and smoothed according to the learned thresholds.
        """
        X_transformed = X.copy()
        for col, thresholds in self.binning_thresholds.items():
            bin_col_name = f'{col}_bin' if not self.replace_vars else col
            binned_data = np.digitize(X_transformed[col].values, thresholds)
            
            # Apply Laplace smoothing to the binned data
            smoothed_binned_data = binned_data + 1  # Add-one smoothing
            
            # Replace or append the smoothed, binned data
            X_transformed[bin_col_name] = smoothed_binned_data
            
            # If replacing, drop the original column if it's not the same as the bin_col_name
            if self.replace_vars and bin_col_name != col:
                X_transformed.drop(col, axis=1, inplace=True)
        
        return X_transformed
################################################################################################



from collections import Counter

import numpy as np
from sklearn.cluster import KMeans


class KMeansFeaturizer:
    """Transforms numeric data into k-means cluster memberships.
    From the Feature Engineering Book by Alice Zheng
    https://github.com/alicezheng/feature-engineering-book/blob/master/kmeans_featurizer.py
    This transformer runs k-means on the input data and converts each data point
    into the id of the closest cluster. If a target variable is present, it is
    scaled and included as input to k-means in order to derive clusters that
    obey the classification boundary as well as group similar points together.
    Parameters
    ----------
    k: integer, optional, default 100
        The number of clusters to group data into.
    target_scale: float, [0, infty], optional, default 5.0
        The scaling factor for the target variable. Set this to zero to ignore
        the target. For classification problems, larger `target_scale` values
        will produce clusters that better respect the class boundary.
    random_state : integer or numpy.RandomState, optional
        This is passed to k-means as the generator used to initialize the
        kmeans centers. If an integer is given, it fixes the seed. Defaults to
        the global numpy random number generator.
    Attributes
    ----------
    cluster_centers_ : array, [k, n_features]
        Coordinates of cluster centers. n_features does count the target column.
    """

    def __init__(self, k=100, target_scale=5.0, random_state=None):
        self.k = k
        self.target_scale = target_scale
        self.random_state = random_state
        self.km_model_ = None
        self.km_model = None
        self.cluster_centers_ = None

    def fit(self, X, y=None):
        """Runs k-means on the input data and find centroids.
        If no target is given (`y` is None) then run vanilla k-means on input
        `X`.
        If target `y` is given, then include the target (weighted by
        `target_scale`) as an extra dimension for k-means clustering. In this
        case, run k-means twice, first with the target, then an extra iteration
        without.
        After fitting, the attribute `cluster_centers_` are set to the k-means
        centroids in the input space represented by `X`.
        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_data_points, n_features)
        y : vector of length n_data_points, optional, default None
            If provided, will be weighted with `target_scale` and included in
            k-means clustering as hint.
        """
        n_features = X.shape[1]
        if y is None:
            # No target variable, just do plain k-means
            km_model = KMeans(n_clusters=self.k,
                              n_init=20,
                              random_state=self.random_state)
            km_model.fit(X)

            self.km_model_ = km_model
            self.cluster_centers_ = km_model.cluster_centers_
            return self

        # There is target information. Apply appropriate scaling and include
        # into input data to k-means
        data_with_target = np.hstack((X, y[:, np.newaxis] * self.target_scale))

        # Build a pre-training k-means model on data and target
        km_model_pretrain = KMeans(n_clusters=self.k,
                                   n_init=20,
                                   random_state=self.random_state)
        km_model_pretrain.fit(data_with_target)

        # Run k-means a second time to get the clusters in the original space
        # without target info. Initialize using centroids found in pre-training.
        # Go through a single iteration of cluster assignment and centroid
        # recomputation.
        km_model = KMeans(n_clusters=self.k,
                          init=km_model_pretrain.cluster_centers_[:, :n_features],
                          n_init=1,
                          max_iter=1)
        km_model.fit(X)

        self.km_model = km_model
        self.cluster_centers_ = km_model.cluster_centers_
        return self

    def transform(self, X):
        """Output the closest cluster id for each input data point.
        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_data_points, n_features)
        Returns
        -------
        cluster_ids : array, shape[n_data_points,1]
        """
        clusters = self.km_model.predict(X)
        return clusters[:, np.newaxis]

    def fit_transform(self, X, y=None):
        """Runs fit followed by transform.
        """
        self.fit(X, y)
        return self.transform(X)


import copy


def Transform_KM_Features(training_data, training_labels, test_data, km_max=0):
    seed = 99
    preds = list(training_data)
    target = training_labels.name
    train_index = training_data.index
    test_index = test_data.index
    if km_max <= 2:
        k_max = 2
    else:
        k_max = copy.deepcopy(km_max)
    ### Calculate the target scale here => the higher the number the better for target accuracy
    try:
        if training_labels.dtype in [np.float64, np.float32, np.float16]:
            target_range = float(abs(training_labels.max() - training_labels.min()))
        elif training_labels.dtype in [object, bool]:
            target_range = int(len(Counter(training_labels)) + 3)
        else:
            target_range = int(abs(training_labels.max() - training_labels.min()))
    except:
        target_range = 5.0
    kmf = KMeansFeaturizer(k=k_max, target_scale=target_range, random_state=seed)
    kmf_hint = kmf.fit(training_data.values, training_labels.values)
    ### Just return it with the cluster column => no need to return the data frame ###
    training_cluster_features = kmf_hint.transform(training_data.values)
    test_cluster_features = kmf_hint.transform(test_data.values)
    npx = np.c_[training_data.values, training_labels.values]
    training_with_cluster = np.c_[npx, training_cluster_features]
    test_with_cluster = np.c_[test_data, test_cluster_features]
    ### We are going to just return the cluster values ######
    train_with_cluster_df = training_with_cluster[:, -1]
    test_with_cluster_df = test_with_cluster[:, -1]
    # train_with_cluster_df = pd.DataFrame(training_with_cluster,index=train_index,
    #                                  columns=preds+[target,'cluster'])
    # test_with_cluster_df = pd.DataFrame(test_with_cluster,index=test_index,
    #                                  columns=preds+['cluster'])
    return train_with_cluster_df, test_with_cluster_df
